from typing import List

from wincam import FpsThrottle, Timer


def test_throttle(test_frames=60, fps=30):
    throttle = FpsThrottle(fps, window_size=1)
    timer = Timer()
    timer.start()
    times: List[float] = []
    previous = None
    for i in range(test_frames):
        timer.sleep(15)
        throttle.step()
        now = timer.ticks()
        if previous is None:
            previous = now
        else:
            times += [now - previous]
            previous = now

    min_step = min(times)
    max_step = max(times)
    avg_step = sum(times) / len(times)

    print(f"frame step times, min: {min_step:.3f}, max: {max_step:.3f}, avg: {avg_step:.3f}")

    with open("times.csv", "w") as f:
        for step in times:
            f.write(f"{step}\n")

    assert len(times) > test_frames * 0.9
    assert len(times) < test_frames * 1.1
    assert avg_step * 1000 > fps * 0.85
    assert avg_step * 1000 < fps * 1.16



import os
from typing import Tuple

import win32gui


class DesktopWindow:
    """This is a helper class that can be used to find window bounds."""

    def __init__(self):
        self.pid = 0
        self.handles = []
        if os.name != "nt":
            raise Exception("This class only works on Windows")

        self.camera = None

    def find(self, pid: int) -> Tuple[int, int, int, int]:
        self.pid = int(pid)
        win32gui.EnumWindows(self._check_window, None)
        hwnd = self._get_visible_window()
        if hwnd == -1:
            raise Exception(f"Visible window not found for process {pid}")

        return self.get_window_client_bounds(hwnd)

    def get_window_under_point(self, point: Tuple[int, int]):
        hwnd = win32gui.WindowFromPoint(point)
        if not hwnd:
            raise Exception(f"Visible window not found at location {point}")
        return self.get_window_client_bounds(hwnd)

    def get_window_client_bounds(self, hwnd):
        rc = win32gui.GetClientRect(hwnd)
        if rc is None:
            raise Exception(f"Window not found {hwnd}")

        left, top, right, bottom = rc
        w = right - left
        h = bottom - top
        left, top = win32gui.ClientToScreen(hwnd, (left, top))
        return (left, top, w, h)

    def _get_visible_window(self):
        best_hwnd = -1
        max_area = 0
        for hwnd in self.handles:
            visible = win32gui.IsWindowVisible(hwnd)
            if not visible:
                continue
            _, _, w, h = self.get_window_client_bounds(hwnd)
            area = w * h
            if w > 0 and h > 0 and area > max_area:
                max_area = area
                best_hwnd = hwnd
        return best_hwnd

    def find_window_by_title(self, title: str) -> int:
        for i in range(len(self.handles)):
            hwnd = self.handles[i]
            text = win32gui.GetWindowText(hwnd)
            if text == title:
                return i
        return -1

    def _check_window(self, hwnd, lparam):
        import win32process

        _, pid = win32process.GetWindowThreadProcessId(hwnd)
        if pid == self.pid:
            self.handles += [hwnd]



from abc import ABC, abstractmethod
from typing import Tuple

import numpy as np

from wincam.logger import Logger

logger = Logger()
log = logger.get_root_logger()


class Camera(ABC):
    """Abstract class for a camera that captures frames from the screen.
    Some cameras may capture frames natively in the BGR format, while others may capture
    in RGB natively.  So one of the two methods will do a conversion to the format you
    require, but the native method will be the fastest."""

    def __init__(self):
        pass

    @abstractmethod
    def get_rgb_frame(self) -> Tuple[np.ndarray, float]:
        """Returns the next RGB image of the screen and the time most closely
        matching when the frame was captured (in seconds)
        starting at 0 for the first call to get_rgb_frame."""
        pass

    def get_bgr_frame(self) -> Tuple[np.ndarray, float]:
        """Returns the next BGR image of the screen and the time most closely
        matching when the frame was captured (in seconds)
        starting at 0 for the first call to get_rgb_frame."""
        pass

    @abstractmethod
    def stop(self):
        pass



import argparse
import ctypes
import json
import os
import signal
from threading import Thread
from typing import List

from common import add_common_args
import cv2

from wincam import DXCamera, Timer


def get_argument_parser():
    """Parse command line arguments."""
    parser = argparse.ArgumentParser("Record an mp4 video of the screen.")
    add_common_args(parser)
    parser.add_argument(
        "--output",
        help="Name of video file to save (default 'video.mp4')",
        default="video.mp4",
    )
    parser.add_argument(
        "--seconds_per_video",
        type=int,
        default=0,
        help="Break up the video into multiple files with this many seconds per video",
    )
    parser.add_argument("--native", help="use GPU provided video encoder", action="store_true")
    return parser


def parse_handle(hwnd) -> int:
    if hwnd.startswith("0x"):
        return int(hwnd[2:], 16)
    return int(hwnd)


class VideoRecorder:
    def __init__(self, output: str = "video.mp4"):
        self._thread: Thread | None = None
        self._monitor: Thread | None = None
        self._stop = False
        self._output = output
        self._video_writer: cv2.VideoWriter | None = None
        signal.signal(signal.SIGINT, self._signal_handler)

    def _signal_handler(self, sig, frame):
        self._stop = True

    def video_thread(self, x: int, y: int, w: int, h: int, fps: int, seconds_per_video: int, native: bool):
        index = 0
        print()
        while not self._stop:
            filename = self._output
            if index > 0:
                filename = os.path.splitext(filename)[0] + f"_{index}.mp4"
            if native:
                self.native_encoder(filename, x, y, w, h, fps, seconds_per_video, index)
            else:
                self.record_video(filename, x, y, w, h, fps, seconds_per_video, index)

            index += 1
            if seconds_per_video == 0:
                break

    def native_encoder(self, filename: str, x: int, y: int, w: int, h: int, fps: int, max_seconds: int, index: int):
        timer = Timer()
        with DXCamera(x, y, w, h, fps=fps) as camera:
            frame, timestamp = camera.get_bgr_frame()
            if index == 0:
                # do a 2 second warm up cycle to ensure video capture is warm
                timer.start()
                while timer.ticks() < 2:
                    frame, timestamp = camera.get_bgr_frame()
            print(f"Recording {filename}...")

            self._monitor = Thread(target=lambda: self.monitor_video(camera, max_seconds))
            self._monitor.start()
            camera.encode_video(filename, 9000000, 60)
            self._monitor.join()
            print("Video saved to", filename)
            ticks = camera.get_video_ticks()
            self.save_video_meta(filename, ticks)
            self.report_steps(self.get_steps(ticks))

    def get_steps(self, ticks: List[float]):
        p = None
        steps: List[float] = []
        for t in ticks:
            if p is None:
                p = t
            else:
                steps.append(t - p)
                p = t
        return steps

    def save_video_meta(self, filename, ticks):
        meta_file = os.path.splitext(filename)[0] + "_meta.json"
        with open(meta_file, "w") as f:
            json.dump({"video_ticks": ticks}, f)

    def monitor_video(self, camera: DXCamera, max_seconds: int):
        timer = Timer()
        timer.start()
        while timer.ticks() < max_seconds and not self._stop:
            timer.sleep(100)
        camera.stop_encoding()

    def record_video(self, filename: str, x: int, y: int, w: int, h: int, fps: int, max_seconds: int, index: int):
        self._video_writer = cv2.VideoWriter(
            filename,
            cv2.VideoWriter_fourcc(*"mp4v"),  # type: ignore
            fps,
            (w, h),
        )
        steps = []
        timer = Timer()
        frame_count = 0
        with DXCamera(x, y, w, h, fps=fps) as camera:
            frame, timestamp = camera.get_bgr_frame()
            self._video_writer.write(frame)  # type: ignore
            camera.reset_throttle()

            if index == 0:
                # do a 2 second warm up cycle to ensure video capture is warm
                timer.start()
                while timer.ticks() < 2:
                    frame, timestamp = camera.get_bgr_frame()

            print(f"Recording {filename}...")
            camera.reset_throttle()
            timer.start()
            step_timer = Timer()

            while not self._stop:
                step_timer.start()
                frame, timestamp = camera.get_bgr_frame()
                self._video_writer.write(frame)
                frame_count += 1
                step_time = step_timer.ticks()
                steps += [step_time]
                if step_time > 1:
                    print(f"frame {frame_count} took {step_time:.3f} seconds??")
                if max_seconds > 0 and timer.ticks() > max_seconds:
                    break

            self._video_writer.release()

        print("Video saved to", filename)
        self.report_steps(steps)

        total = timer.ticks()
        avg_fps = frame_count / total
        print(f"Recorded {frame_count} frames at average fps {avg_fps}")

        if index == 0 and avg_fps < fps * 0.9:
            print(f"The video writer could not keep up with the target {fps} fps so the video will play too fast.")
            print("Please try a smaller window or a lower target fps.")

    def report_steps(self, ticks):
        if len(ticks) > 0:
            min_step = min(ticks)
            max_step = max(ticks)
            avg_step = sum(ticks) / len(ticks)
            print(f"frame step times, min: {min_step:.3f}, max: {max_step:.3f}, avg: {avg_step:.3f}")

    def start(self, x: int, y: int, w: int, h: int, fps: int, seconds_per_video: int, native: bool):
        self._thread = Thread(target=lambda: self.video_thread(x, y, w, h, fps, seconds_per_video, native))
        self._thread.start()

    def stop(self):
        self._stop = True
        if self._thread is not None:
            self._thread.join()


def main():
    ctypes.windll.shcore.SetProcessDpiAwareness(2)
    parser = get_argument_parser()
    args = parser.parse_args()
    if args.debug:
        input(f"debug process {os.getpid()}")

    x, y, w, h = args.x, args.y, args.width, args.height
    if args.hwnd:
        hwnd = parse_handle(args.hwnd)
        from wincam.desktop import DesktopWindow

        desktop = DesktopWindow()
        x, y, w, h = desktop.get_window_client_bounds(hwnd)
    elif args.point:
        x, y = map(int, args.point.split(","))
        from wincam.desktop import DesktopWindow

        desktop = DesktopWindow()
        x, y, w, h = desktop.get_window_under_point((x, y))
    elif args.process:
        pid = int(args.process)
        from wincam.desktop import DesktopWindow

        desktop = DesktopWindow()
        x, y, w, h = desktop.find(pid)

    recorder = VideoRecorder(args.output)
    recorder.start(x, y, w, h, args.fps, args.seconds_per_video, args.native)
    input("Press ENTER to stop recording...")
    recorder.stop()


if __name__ == "__main__":
    try:
        main()
    except Exception as e:
        print(f"Error: {e}")



import logging

PREFIX = "wincam"

has_console_handler = False


class Logger:
    """Logger class to be used by all modules in the project.

    The available logging levels are given below in decreasing order of severity:
        CRITICAL
        ERROR
        WARNING
        INFO
        DEBUG

    When a log_level is set to a logger, logging messages which are less severe than it will be ignored.
    For example, if the log_level is set to INFO, then DEBUG messages will be ignored.
    """

    def __init__(self):
        self.root_logger = logging.getLogger(PREFIX)
        self.log_level = "INFO"

    def get_root_logger(self, log_level="INFO", log_file=None):
        self._setup(log_level, log_file)
        return self.root_logger

    def set_log_level(self, log_level):
        self.log_level = log_level
        self.root_logger.setLevel(log_level)
        for handler in self.root_logger.handlers:
            handler.setLevel(log_level)

    def set_log_file(self, log_file):
        self._add_file_handler(log_file)

    def _setup(self, log_level, log_file):
        self.log_level = log_level
        self.root_logger.setLevel(log_level)
        self._add_console_handler()
        if log_file:
            self._add_file_handler(log_file)

        # To avoid having duplicate logs in the console
        self.root_logger.propagate = False

    def _add_console_handler(self):
        global has_console_handler
        if not has_console_handler:
            console_handler_formatter = logging.Formatter("%(filename)s [%(levelname)s]: %(message)s")
            console_handler = logging.StreamHandler()
            console_handler.setLevel(self.log_level)
            console_handler.setFormatter(console_handler_formatter)
            self.root_logger.addHandler(console_handler)
            has_console_handler = True

    def _add_file_handler(self, log_file):
        file_handler_formatter = logging.Formatter(
            "%(asctime)s %(filename)s [%(levelname)s]: %(message)s", datefmt="%Y-%m-%d %H:%M:%S"
        )
        file_handler = logging.FileHandler(log_file)
        file_handler.setLevel(self.log_level)
        file_handler.setFormatter(file_handler_formatter)
        self.root_logger.addHandler(file_handler)

    @staticmethod
    def get_logger(name):
        # We enforce the creation of a child logger (PREFIX.name) to keep the root logger setup
        if name.startswith(PREFIX + "."):
            return logging.getLogger(name)
        return logging.getLogger(PREFIX + "." + name)



import argparse
import ctypes
import os

from common import add_common_args
import cv2

from wincam import DXCamera


def get_argument_parser():
    """Parse command line arguments."""
    parser = argparse.ArgumentParser("Mirror a region of the screen.")
    add_common_args(parser)
    return parser


def parse_handle(hwnd) -> int:
    if hwnd.startswith("0x"):
        return int(hwnd[2:], 16)
    return int(hwnd)


def main():
    ctypes.windll.shcore.SetProcessDpiAwareness(2)
    parser = get_argument_parser()
    args = parser.parse_args()
    print("Press ESCAPE to close the window...")
    if args.debug:
        input(f"debug process {os.getpid()}")

    x, y, w, h = args.x, args.y, args.width, args.height
    if args.hwnd:
        hwnd = parse_handle(args.hwnd)
        from wincam.desktop import DesktopWindow

        desktop = DesktopWindow()
        x, y, w, h = desktop.get_window_client_bounds(hwnd)
    elif args.point:
        x, y = map(int, args.point.split(","))
        from wincam.desktop import DesktopWindow

        desktop = DesktopWindow()
        x, y, w, h = desktop.get_window_under_point((x, y))
    elif args.process:
        pid = int(args.process)
        from wincam.desktop import DesktopWindow

        desktop = DesktopWindow()
        x, y, w, h = desktop.find(pid)

    with DXCamera(x, y, w, h, args.fps) as camera:
        while True:
            frame, timestamp = camera.get_bgr_frame()
            cv2.imshow("frame", frame)
            if cv2.waitKey(1) & 0xFF == 27:
                break


if __name__ == "__main__":
    try:
        main()
    except Exception as e:
        print(f"Error: {e}")



from wincam.camera import Camera
from wincam.dxcam import DXCamera
from wincam.logger import Logger
from wincam.throttle import FpsThrottle
from wincam.timer import Timer

__all__ = ["Camera", "DXCamera", "Logger", "Timer", "FpsThrottle"]



import os
import time


class Timer:
    """This class provides accurate timer for recording when user inputs happen
    so that the playback of those events can synchronize the playback to run in the
    same time as the recording."""

    WINDOWS_SLEEP_ACCURACY_MS = 15

    def __init__(self):
        self._start_time = None

    def start(self):
        """Save the start time"""
        self._start_time = time.perf_counter()

    def ticks(self) -> float:
        """Returns 0 on the first call and the delta from this first call in seconds
        on subsequent calls using time.perf_counter() for higher precision."""
        if self._start_time is None:
            self.start()
            return 0
        return time.perf_counter() - self._start_time

    def sleep(self, milliseconds: int):
        """Sleep for the given number of milliseconds."""
        if os.name == "nt":
            self._windows_sleep(milliseconds)
        else:
            self._unix_sleep(milliseconds)

    def _windows_sleep(self, milliseconds: int):
        # On windows the Win32 Sleep function has an accuracy of about 15ms which is not good enough
        # so we sleep up to without that tolerance, then spin wait the rest to hit our target.
        now = self.ticks()
        if milliseconds > Timer.WINDOWS_SLEEP_ACCURACY_MS:
            time.sleep(float(milliseconds - Timer.WINDOWS_SLEEP_ACCURACY_MS) / 1000.0)

        # spin wait for the rest.
        while (self.ticks() - now) * 1000 < milliseconds:
            pass

    def _unix_sleep(self, milliseconds: int):
        """Unix sleep call."""
        time.sleep(float(milliseconds) / 1000.0)



import ctypes as ct
import os
from typing import Tuple

import cv2
import numpy as np

from wincam.camera import Camera
from wincam.throttle import FpsThrottle

script_dir = os.path.dirname(os.path.realpath(__file__))


class Rect(ct.Structure):
    _fields_ = [("x", ct.c_int), ("y", ct.c_int), ("width", ct.c_int), ("height", ct.c_int)]


class DXCamera(Camera):
    _instance = None

    """ Camera that captures frames from the screen using the ScreenCapture.dll native library
    which is based on Direct3D11CaptureFramePool.
    See https://learn.microsoft.com/en-us/uwp/api/windows.graphics.capture.direct3d11captureframepool"""

    def __init__(self, left: int, top: int, width: int, height: int, fps: int = 30, capture_cursor: bool = True):
        super().__init__()
        if os.name != "nt":
            raise Exception("This class only works on Windows")

        self._width = width
        self._height = height
        self._left = left
        self._top = top
        self._capture_cursor = capture_cursor
        self._throttle = FpsThrottle(fps)
        full_path = os.path.realpath(os.path.join(script_dir, "native", "runtimes", "x64", "ScreenCapture.dll"))
        if not os.path.exists(full_path):
            raise Exception(f"ScreenCapture.dll not found at: {full_path}")
        self.lib = ct.cdll.LoadLibrary(full_path)
        self.lib.GetCaptureBounds.restype = Rect
        self.lib.EncodeVideo.argtypes = [ct.c_uint32, ct.c_wchar_p, ct.c_int, ct.c_int]
        self.lib.EncodeVideo.restype = ct.c_uint32
        self.lib.GetTicks.argtypes = [ct.POINTER(ct.c_double), ct.c_int]
        self.lib.GetTicks.restype = ct.c_uint32
        self._started = False
        self._buffer = None
        self._size = 0
        self._capture_bounds = Rect()
        self._handle = -1

    def __enter__(self):
        if self._instance is None:
            DXCamera._instance = self
        else:
            raise Exception("You can only use 1 instance of DXCamera at a time.")
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        DXCamera._instance = None
        self.stop()

    def reset_throttle(self):
        self._throttle.reset()

    def get_bgr_frame(self) -> Tuple[np.ndarray, float]:
        if not self._started:
            self._handle = self.lib.StartCapture(self._left, self._top, self._width, self._height, self._capture_cursor)
            if not self.lib.WaitForNextFrame(self._handle, 10000):
                raise Exception("Frames are not being captured")

            self._capture_bounds = self.lib.GetCaptureBounds(self._handle)
            self._size = self._capture_bounds.width * self._capture_bounds.height * 4
            self._buffer = ct.create_string_buffer(self._size)  # type: ignore
            self._started = True
            self._throttle.reset()

        timestamp = self.lib.ReadNextFrame(self._handle, self._buffer, len(self._buffer))
        image = np.resize(
            np.frombuffer(self._buffer, dtype=np.uint8), (self._capture_bounds.height, self._capture_bounds.width, 4)
        )
        # strip out the alpha channel, and any 64-byte aligned extra width
        image = image[:, : self._width, :3]

        self._throttle.step()
        return image, timestamp / 1000000.0

    def get_rgb_frame(self) -> Tuple[np.ndarray, float]:
        frame, timestamp = self.get_bgr_frame()
        frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
        return frame, timestamp

    def encode_video(self, file_name: str, bit_rate: int = 9000000, frame_rate: int = 60):
        self.get_bgr_frame()  # make sure we're getting frames.
        full_path = os.path.realpath(file_name)
        if os.path.isfile(full_path):
            os.remove(full_path)
        self.lib.EncodeVideo(self._handle, full_path, bit_rate, frame_rate)

    def stop_encoding(self):
        self.lib.StopEncoding()

    def get_video_ticks(self):
        len = self.lib.GetTicks(None, 0)
        if len > 0:
            array = (ct.c_double * len)()
            self.lib.GetTicks(array, len)
            return list(array)
        return []

    def stop(self):
        self._started = False
        self.lib.StopCapture(self._handle)
        self.lib.StopEncoding()
        self._buffer = None
        self._handle = -1



def add_common_args(parser):
    parser.add_argument(
        "--fps",
        type=int,
        help="Desired frame rate for recording (default 30)",
        default=30,
    )
    parser.add_argument(
        "--x",
        type=int,
        help="Desired X origin of the mirror (default 0)",
        default=0,
    )
    parser.add_argument(
        "--y",
        type=int,
        help="Desired Y origin of the mirror (default 0)",
        default=0,
    )
    parser.add_argument(
        "--width",
        type=int,
        help="Desired width of the mirror (default 1024)",
        default=1024,
    )
    parser.add_argument(
        "--height",
        type=int,
        help="Desired height of the mirror (default 720)",
        default=720,
    )
    parser.add_argument(
        "--hwnd",
        help="Desired window handle to record.  You can find these using the Windows Kit tool named 'inspect'."
        + "Support hex format starting with 0x.",
    )
    parser.add_argument(
        "--point",
        help="Find window under the given point provided as 'x,y'",
    )
    parser.add_argument(
        "--process",
        help="Find window belonging to the given process id.",
    )
    parser.add_argument(
        "--debug",
        help="For debugging only",
        action="store_true",
    )



import numpy as np

from wincam.timer import Timer


class FpsThrottle:
    """Helper class that throttles the frame rate to a given fps. Simply set the frame rate in the constructor then call
    step and step will sleep the right amount to make the time between step calls equal the given fps.  In order to
    avoid large initial times it smooths this time only over the window_size (in seconds) so it settles into a nice
    steady state more quickly than it would if it was taking the overall average."""

    def __init__(self, fps: int, window_size=10):
        self.fps = fps
        self.window_size = window_size
        self.reset()
        self.target_ms_per_frame = 1000.0 / self.fps
        self.window: np.ndarray | None = None
        self.pos = 0

    def reset(self):
        self.timer = Timer()
        self.started = False
        self.window = None
        self.pos = 0

    def step(self):
        if not self.started:
            self.started = True
            throttle_ms = 0.0
        else:
            # get time for full step loop since we were last called
            average_ms_per_frame = self.timer.ticks() * 1000.0
            if self.window is None:
                self.window = np.ones((self.window_size,)) * average_ms_per_frame
            else:
                self.window[self.pos] = average_ms_per_frame
                self.pos = (self.pos + 1) % self.window_size
            smoothed_average = np.mean(self.window)
            throttle_ms = self.target_ms_per_frame - smoothed_average

        # sync to match the given frame rate.
        if throttle_ms > 0:
            self.timer.sleep(int(throttle_ms))
        self.timer.start()
        return throttle_ms
